"""
## Development

### release new version

    git commit -am "version bump";git push origin master
    python setup.py --version
    git tag -a v$(python setup.py --version) -m "upgrade";git push --tags

"""

import sys
if (sys.version_info[0]) != (3):
     raise RuntimeError('Python 3 required ')

import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

## dependencies/requirements
requirements = {
'base':  """biopython>=1.71
            regex>=2018.7.11
            pandas>=0.25.3
            pyyaml>=5.1
            numpy>=1.17.3
            matplotlib>=2.2
            requests>=2.19.1
            scipy>=1.1.0
            tqdm>=4.38.0
            fastparquet
            pandarallel
            xlrd
            openpyxl
            argh
            icecream
            watermark""".split('\n'),
'stat':['statsmodels',
        'networkx',
        # 'pingouin',
        ],
'viz':['seaborn>=0.8',
      ],
'query':['pyensembl',
         'pybiomart'],
'workflow':['snakemake',
            'gitpython',
            'networkx',
            'pygraphviz'],
'dev':[
    'pytest',
    'ipython','ipywidgets','jupyter',
    'sphinx',
    'recommonmark',
    'black',
    'coveralls == 3.*',
    'flake8',
    'isort',
    'pytest-cov == 2.*',
    'testbook',
],
}
extras_require={k:l for k,l in requirements.items() if not k=='base'}
## all: extra except dev
extras_require['all']=[l for k,l in extras_require.items() if not k=='dev']
### flatten
extras_require['all']=[s for l in extras_require['all'] for s in l]
### unique
extras_require['all']=list(set(extras_require['all']))

setuptools.setup(
    name='roux',
    version='0.0.6',
    description='helper functions',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='http://github.com/rraadd88/roux',
    author='rraadd88',
    author_email='rohanadandage@gmail.com',
    license='General Public License v. 3',
    packages=setuptools.find_packages('.',exclude=['test', 'unit','deps', 'data']),
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=requirements['base'],
    extras_require=extras_require,
    entry_points={
    'console_scripts': ['roux = roux.run:parser.dispatch',],
    },
    python_requires='>=3.7, <4',
)



import pandas as pd
import numpy as np
import scipy as sc
import logging

def plog(x, p: float, base: int):
    """Psudo-log.

    Args:
        x (float|np.array): input.
        p (float): pseudo-count.
        base (int): base of the log.

    Returns:
        output.
    """
    if not base is None:
        return np.log(x+p)/np.log(base)
    else:
        return np.log(x+p)

def anti_plog(x, p: float, base: int):
    """Anti-psudo-log.

    Args:
        x (float|np.array): input.
        p (float): pseudo-count.
        base (int): base of the log.

    Returns:
        output.
    """
    return (base**x)-p
    
def log_pval(x):
    """Transform p-values to Log10.

    Paramters: 
        x: input.

    Returns:
        output.
    """ 
    if not isinstance(x,(int,float)): 
        if any(x==0):
            x=x.replace(0,x.replace(0,np.nan).min())
            logging.warning('zeros found, replaced with min pval')
    return -1*(np.log10(x))

def glog(x: float,l = 2):
    """Generalised logarithm.

    Args:
        x (float): input.
        l (int, optional): psudo-count. Defaults to 2.

    Returns:
        float: output.
    """
    return np.log((x+np.sqrt(x**2+l**2))/2)/np.log(l)

def rescale(a: np.array, range1: tuple=None, range2: tuple=[0,1]) -> np.array:
    """Rescale within a new range.

    Args:
        a (np.array): input vector.
        range1 (tuple, optional): existing range. Defaults to None.
        range2 (tuple, optional): new range. Defaults to [0,1].

    Returns:
        np.array: output.
    """
    if range1 is None:
        range1=[np.min(a),np.max(a)]
    delta1 = range1[1] - range1[0]
    delta2 = range2[1] - range2[0]
    return (delta2 * (a - range1[0]) / delta1) + range2[0]

def rescale_divergent(df1: pd.DataFrame,col: str,
#                       rank=True,
                     ) -> pd.DataFrame:
    """Rescale divergently i.e. two-sided.

    Args:
        df1 (pd.DataFrame): _description_
        col (str): column.

    Returns:
        pd.DataFrame: column.

    Notes:
        Under development.
    """
    def apply_(df2,
#                sign=None,
              ):
        sign=df2.name
#         from roux.stat.transform import rescale
        df2[f'{col} rescaled']=rescale(df2[col],range2=[0, 1] if sign=='+' else [-1,0])
        df2[f'{col} rank']=df2[col].rank(ascending=True if sign=='+' else False)*(1 if sign=='+' else -1)
        return df2
    assert(not any(df1[col]==0))
    df1.loc[df1[col]<0,f'{col} sign']='-'
    df1.loc[df1[col]>0,f'{col} sign']='+'
#     return pd.concat({k:apply_(df1.loc[(df1[f'{col} sign']==k),:],k) for k in df1[f'{col} sign'].unique()},
#                     axis=0)
    return df1.groupby([f'{col} sign']).apply(lambda df: apply_(df))


#!usr/bin/python

"""Strings."""
import re
import logging
import numpy as np

# convert
def substitution(s,i,replaceby):
    """Substitute character in a string.
    
    Parameters:
        s (string): string.
        i (int): location.
        replaceby (string): character to substitute with.
        
    Returns:
        s (string): output string.
    """
    l=list(s)
    l[i]=replaceby
    return "".join(l)
# alias
replacebyposition=substitution

def replace_many(s,replaces,replacewith='',
               ignore=False):
    """Rename by replacing sub-strings.

    Parameters:
        s (str): input string.
        replaces (dict|list): from->to format or list containing substrings to remove.
        replacewith (str): replace to in case `replaces` is a list.
        ignore (bool): if True, not validate the successful replacements.
        
    Returns:
        s (DataFrame): output dataframe.
    """
    s_=s
    if isinstance(replaces,list):
        replaces={k:replacewith for k in replaces}
    if isinstance(replaces,dict):
        for k in replaces:
            s=s.replace(k,replaces[k])
    else:
        import inspect
        if inspect.isfunction(replaces):
            s=replaces(s_)
        else:
            raise ValueError(replaces)
    if not ignore: assert(s!=s_)
    return s
# alias
replacemany=replace_many

def tuple2str(tup,sep=' '):
    """Join tuple items.
    
    Parameters:
        tup (tuple|list): input tuple/list.
        sep (str): separator between the items.
        
    Returns:
        s (str): output string.
    """    
    if isinstance(tup,tuple):
        tup=[str(s) for s in tup if not s=='']
        if len(tup)!=1:
            tup=sep.join(list(tup))
        else:
            tup=tup[0]
    elif not isinstance(tup,str):
        logging.error("tup is not str either")
    return tup

# def normalisestr(s,):
#     """Normalise string.
    
#     Parameters:
#         s (string): input string.
    
#     Returns:
#         s (string): output string.
#     """
#     if not isinstance(s,str):
#         s=s.decode("utf-8")
#     import re
#     return re.sub('\W+','', s.lower()).replace('_','')

def linebreaker(text, 
                width=None,
                break_pt=None,
                sep='\n',
                **kws,
               ):
    """Insert `newline`s within a string. 
    
    Parameters:
        text (str): string.
        width (int): insert `newline` at this interval. 
        sep (string): separator to split the sub-strings.
        
    Returns:
        s (string): output string.
        
    References:
        1. `textwrap`: https://docs.python.org/3/library/textwrap.html
    """
    if width is None and not break_pt is None:
        width=break_pt
    import textwrap
    return sep.join(textwrap.wrap(text, width,**kws,))
    # if len(i)>break_pt:
    #     i_words=i.split(sep)
    #     i_out=''
    #     line_len=0
    #     for w in i_words:
    #         line_len+=len(w)+1
    #         if i_words.index(w)==0:
    #             i_out=w
    #         elif line_len>break_pt:
    #             line_len=0
    #             i_out="%s\n%s" % (i_out,w)
    #         else:
    #             i_out="%s %s" % (i_out,w)    
    #     return i_out    
    # else:
    #     return i

# find
def findall(s,ss,outends=False,outstrs=False,
           suffixlen=0):
    """Find the substrings or their locations in a string.
   
    Parameters:
        s (string): input string.
        ss (string): substring.
        outends (bool): output end positions.
        outstrs (bool): output strings.
        suffixlen (int): length of the suffix.
    
    Returns:
        l (list): output list.
    """
    import re
    finds=list(re.finditer(ss, s))
    if outends or outstrs:
        locs=[(a.start(), a.end()) for a in finds]
        if not outstrs:
            return locs
        else:
            return [s[l[0]:l[1]+suffixlen] for l in locs]
    else:
        return [a.start() for a in finds]
        
def get_marked_substrings(s,leftmarker='{',rightmarker='}',
                          leftoff=0,rightoff=0):
    """Get the substrings flanked with markers from a string.
    
    Parameters:
        s (str): input string.
        leftmarker (str): marker on the left.
        rightmarker (str): marker on the right. 
        leftoff (int): offset on the left.
        rightoff (int): offset on the right.
    
    Returns:
        l (list): list of substrings.
    """
    filers=[]
    for ini, end in zip(findall(s,leftmarker,outends=False),findall(s,rightmarker,outends=False)):
        filers.append(s[ini+1+leftoff:end+rightoff])
    return filers    

getall_fillers=get_marked_substrings
###
def mark_substrings(s,ss,
                  leftmarker='(',rightmarker=')'):
    """Mark sub-string/s in a string.
    
    Parameters:
        s (str): input string.
        ss (str): substring.
        leftmarker (str): marker on the left.
        rightmarker (str): marker on the right. 
    
    Returns:
        s (str): string.
    """
    pos=s.find(ss)
    return f"{s[:pos]}{leftmarker}{s[pos:pos+len(ss)]}{rightmarker}"

def get_bracket(s,leftmarker='(',righttmarker=')'):
    """Get bracketed substrings.
    
    Parameters:
        s (string): string.
        leftmarker (str): marker on the left.
        rightmarker (str): marker on the right. 
    
    Returns:
        s (str): string.
        
    TODOs:
        1. Use `get_marked_substrings`.
    """
#     import re
#     re.search(r'{l}(.*?){r}', s).group(1)    
    if leftmarker in s and righttmarker in s:
        return s[s.find(leftmarker)+1:s.find(righttmarker)]
    else:
        return '' 
    
## split
def align(s1,s2,
          prefix=False,
          suffix=False,
          common=True):
    """Align strings.
        
    Parameters:
        s1 (str): string #1.
        s2 (str): string #2.
        prefix (str): prefix.
        suffix (str): suffix.
        common (str): common substring.
    
    Returns:
        l (list): output list.
        
    Notes:
        1. Code to test:
            [
            get_prefix(source,target,common=False),
            get_prefix(source,target,common=True),
            get_suffix(source,target,common=False),
            get_suffix(source,target,common=True),]
    """
    
    for i,t in enumerate(zip(list(s1),list(s2))):
        if t[0]!=t[1]:
            break
    if common:
        return [s1[:i],s2[:i]] if prefix else [s1[i+1:],s2[i+1:]] 
    else:
        return [s1[:i+1],s2[:i+1]] if prefix else [s1[i:],s2[i:]] 

from roux.lib.set import unique_str    
def get_prefix(s1,s2,common=True,clean=True): 
    """Get the prefix of the strings
    
    Parameters:
        s1 (str): 1st string.
        s2 (str): 2nd string.
        common (bool): get the common prefix (default:True).
        clean (bool): clean the leading and trailing whitespaces (default:True).
        
    Returns:
        s (str): prefix.
    """
    l1=align(s1,s2,prefix=True,common=common)
    if not common:
        return l1
    else:
        s3=unique_str(l1)
        if not clean:
            return s3
        else:
            return s3.strip().rsplit(' ', 1)[0]    
def get_suffix(s1,s2,common=True,clean=True):
    """Get the suffix of the strings
    
    Parameters:
        s1 (str): 1st string.
        s2 (str): 2nd string.
        common (bool): get the common prefix (default:True).
        clean (bool): clean the leading and trailing whitespaces (default:True).
        
    Returns:
        s (str): suffix.
    """    
    l1=align(s1,s2,suffix=True,common=common)
    if not common:
        if not clean:
            return l1
        else:
            split_pos=(max([s.count(' ') for s in l1])+1)*-1
            return [' '.join(s.split(' ')[split_pos:]) for s in [s1,s2]]
    else:
        s3=unique_str(l1)
        if not clean:
            return s3
        else:
            return s3.strip()#.rsplit(' ', 1)[0]
def get_fix(s1,s2,**kws):
    """Infer common prefix or suffix.
    
    Parameters:
        s1 (str): 1st string.
        s2 (str): 2nd string.
    
    Keyword parameters:
        kws: parameters provided to the `get_prefix` and `get_suffix` functions.
        
    Returns:
        s (str): prefix or suffix.
    """
    s3=get_prefix(s1,s2,**kws)
    s4=get_suffix(s1,s2,**kws)
    return s3 if len(s3)>=len(s4) else s4

def removesuffix(s1,suffix):
    """Remove suffix.
    
    Paramters:
        s1 (str): input string.
        suffix (str): suffix.
        
    Returns:
        s1 (str): string without the suffix.
        
    TODOs: 
        1. Deprecate in py>39 use .removesuffix() instead.
    """
    if s1.endswith(suffix):
        return s1[:s1.rfind(suffix)]
    else:
        return s1

# dict
# def str2dict(s): return dict(item.split("=") for item in s.split(";"))
def str2dict(s,sep=';',sep_equal='='):
    """String to dictionary.
    
    Parameters:
        s (str): string.
        sep (str): separator between entries (default:';').
        sep_equal (str): separator between the keys and the values (default:'=').
    
    Returns:
        d (dict): dictionary.
    
    References:
        1. https://stackoverflow.com/a/186873/3521099
    """
    return dict(item.split(sep_equal) for item in s.split(sep))

def dict2str(d1,sep=';',sep_equal='='): 
    """Dictionary to string.
    
    Parameters:
        d (dict): dictionary.
        sep (str): separator between entries (default:';').
        sep_equal (str): separator between the keys and the values (default:'=').
    
    Returns:
        s (str): string.
    """
    return sep.join([sep_equal.join([k,str(v)]) for k,v in d1.items()])

def str2num(s):
    """String to number.
    
    Parameters:
        s (str): string.
        
    Returns:
        i (int): number.
    """
    import re
    s1=" ".join(re.findall("[a-zA-Z]+", s))
    assert len(s1)==1, f"len({s1})!=1"
    assert s1==s[-1], f"not at the end"
    i1=" ".join(re.findall("[0-9]+", s))
    assert len(s)==len(s1)+len(i1), f"do not add up"
    return int(int(i1)*{'':1, 'K':1e3, 'M':1e6, 'G':1e9, 'T':1e12, 'P':1e15}[s1])

def num2str(num,magnitude=False,
           coff=10000,decimals=0):
    """Number to string.
    
    Parameters:
        num (int): number.
        magnitude (bool): use magnitudes (default:False).
        coff (int): cutoff (default:10000).
        decimals (int): decimal points (default:0).
        
    Returns:
        s (str): string.
        
    TODOs
        1. ~ if magnitude else not
    """
    if not magnitude:
        return f"{num:.1e}" if num>coff else f"{num}"
    else:
        magnitude = 0
        while abs(num) >= 1000:
            magnitude += 1
            num /= 1000.0
        # add more suffixes if you need them
        if decimals==0:
#             return '%.0f%s' % (num, ['', 'K', 'M', 'G', 'T', 'P'][magnitude])
            return '%.0f%s' % (num, ['', 'K', 'M', 'G', 'T', 'P'][magnitude])
        elif decimals==1:
#             return ('%.1f%s' % (num, ['', 'K', 'M', 'G', 'T', 'P'][magnitude])).replace('.0','')
            return ('%.1f%s' % (num, ['', 'K', 'M', 'G', 'T', 'P'][magnitude])).replace('.0','')

## ids
def encode(data,**kws):
    """Encode the data as a string.
    
    Parameters:
        data (dict|Series): input data. 
        
    Keyword parameters:
        kws: parameters provided to `encode`.
        
    Returns:
        s (string): output string. 
    """
    import pandas as pd
    if isinstance(data,pd.Series):
        data=data.to_dict()
    if isinstance(data,dict):
#         from roux.lib.str import dict2str
        data=dict2str(dict(sorted(data.items())), sep=';', sep_equal='=')
    if not isinstance(data,bytes):
        data=data.encode(encoding='utf8')
    import zlib
    from base64 import urlsafe_b64encode as b64e
    return b64e(zlib.compress(data, 9)).decode("utf-8",**kws)

def decode(s,out=None,**kws):
    """Decode data from a string.
    
    Parameters:
        s (string): encoded string. 
        out (str): output format (dict|df).
        
    Keyword parameters:
        kws: parameters provided to `dict2df`.
        
    Returns:
        d (dict|DataFrame): output data. 
    """
    import zlib
    from base64 import urlsafe_b64decode as b64d
    s2=zlib.decompress(b64d(s)).decode("utf-8")
    if out in ['dict','df']:
#         from roux.lib.str import str2dict
        d1=str2dict(s2, sep=';', sep_equal='=')
        if out=='dict':
            return d1
        elif out=='df':
            from roux.lib.df import dict2df
            return dict2df(d1,**kws)
    else:
        return s2


### Ensembl
# for human genome
# release 77 uses human reference genome GRCh38
# from pyensembl import EnsemblRelease
# EnsemblRelease(release=100)
# for many other species
# ensembl = pyensembl.EnsemblRelease(species=pyensembl.species.Species.register(
# latin_name='saccharomyces_cerevisiae',
# synonyms=['saccharomyces_cerevisiae'],
# reference_assemblies={
#     'R64-1-1': (92, 92),
# }),release=92)
from roux.lib.df import *
import numpy as np
import pandas as pd
import logging
import requests, sys

release2prefix={100:'apr2020.archive',
                101:'aug2020.archive',
                102:'nov2020.archive',
                103:'feb2021.archive',
                104:'may2021.archive',
                 93:'jul2018.archive',
                 75:'feb2014.archive',
                 'grch37':'grch37',
                }
## Ref: https://m.ensembl.org/info/website/archives/index.html 


#pyensembl faster
def to_gene_name(k: str ,ensembl: object) -> str:
    """Gene id to gene name.

    Args:
        k (str): gene id.
        ensembl (object): ensembl object.

    Returns:
        str: gene name.

    Notes:
        1. `ensembl` object.
            from pyensembl import EnsemblRelease
            ensembl EnsemblRelease(release=100)    
    """
    try:
        return ensembl.gene_name_of_gene_id(k)
    except:
        return np.nan
    
def to_protein_id(k: str,ensembl: object) -> str:
    """Transcript id to protein id.

    Args:
        x (str): transcript id.
        ensembl (str): ensembl object.

    Returns:
        str: protein id.
        
    Notes:
        1. `ensembl` object.
            from pyensembl import EnsemblRelease
            ensembl EnsemblRelease(release=100)    
    """
    try:
        t=ensembl.transcript_by_id(k)
        return t.protein_id
    except:
        return np.nan    
    
def to_gene_id(k: str ,ensembl: object) -> str:
    """Transcript id to gene id.

    Args:
        k (str): transcript id.
        ensembl (object): ensembl object.

    Returns:
        str: gene id.
        
    Notes:
        1. `ensembl` object.
            from pyensembl import EnsemblRelease
            ensembl EnsemblRelease(release=100)    
    """
    try:
        t=ensembl.transcript_by_id(k)
        return t.gene_id
    except:
        return np.nan 
    
def to_transcript_id(k: str ,ensembl: object) -> str:
    """Protein id to transcript id.

    Args:
        k (str): protein id.
        ensembl (object): ensembl object.

    Returns:
        str: transcript id.
        
    Notes:
        1. `ensembl` object.
            from pyensembl import EnsemblRelease
            ensembl EnsemblRelease(release=100)    
    """
    try:
        return ensembl.transcript_id_of_protein_id(k)
    except:
        return np.nan 
    
def to_dnaseq(k: str ,ensembl: object) -> str:
    """Gene id to DNA sequence.

    Args:
        k (str): gene id.
        ensembl (object): ensembl object.

    Returns:
        str: DNA sequence.
        
    Notes:
        1. `ensembl` object.
            from pyensembl import EnsemblRelease
            ensembl EnsemblRelease(release=100)    
    """
    try:
        g=ensembl.gene_by_id(k)
        ts=g.transcripts
        lens=[len(t.protein_sequence) if not t.protein_sequence is None else 0 for t in ts]
        return ts[lens.index(max(lens))].id, ts[lens.index(max(lens))].protein_sequence
    except:
        return np.nan,np.nan    
    
def to_protein_id_longest(k: str ,ensembl: object) -> str:
    """Gene id to protein id of the longest protein.

    Args:
        k (str): gene id.
        ensembl (object): ensembl object.

    Returns:
        str: protein id.
        
    Notes:
        1. `ensembl` object.
            from pyensembl import EnsemblRelease
            ensembl EnsemblRelease(release=100)    
    """    
    g=ensembl.gene_by_id(gid)
    return pd.Series({t.protein_id:t.protein_sequence for t in g.transcripts}).dropna().apply(len).sort_values().tail(1).index.to_list()[0]

def to_protein_seq(k: str ,ensembl: object,
                    transcript: bool=False) -> str:
    """Protein/transcript id to protein sequence.

    Args:
        k (str): protein id.
        ensembl (object): ensembl object.

    Returns:
        str: protein sequence.
        
    Notes:
        1. `ensembl` object.
            from pyensembl import EnsemblRelease
            ensembl EnsemblRelease(release=100)    
    """    
    try:
        # t=ensembl.transcript_by_id(k)
        # return t.protein_sequence
        t=ensembl.protein_sequence(k)
        if not length:
            return t
        else:
            return len(t)            
    except:
        return np.nan    
    
def to_cdsseq(k: str,
               ensembl: object) -> str:
    """Transcript id to coding sequence (CDS).

    Args:
        k (str): transcript id.
        ensembl (object): ensembl object.

    Returns:
        str: CDS sequence.
        
    Notes:
        1. `ensembl` object.
            from pyensembl import EnsemblRelease
            ensembl EnsemblRelease(release=100)    
    """    
    try:
        t=ensembl.transcript_by_id(k)
        return t.coding_sequence
    except:
        return np.nan 
    
def get_utr_sequence(k: str,ensembl: object,loc: str='five') -> str:
    """Protein id to UTR sequence.

    Args:
        k (str): transcript id.
        ensembl (object): ensembl object.
        loc (str): location of the UTR.

    Returns:
        str: UTR sequence.
        
    Notes:
        1. `ensembl` object.
            from pyensembl import EnsemblRelease
            ensembl EnsemblRelease(release=100)    
    """    
    try:
        t=ensembl.transcript_by_protein_id(k)
        return getattr(t,f'{loc}_prime_utr_sequence')
    except: 
        logging.warning(f"{k}: no sequence found")
        return     
    
def is_protein_coding(k: str,ensembl: object,geneid: bool=True) -> bool:
    """A gene or protein is protein coding or not.

    Args:
        k (str): protein/gene id.
        ensembl (object): ensembl object.
        geneid (bool): if gene id is provided.

    Returns:
        bool: is protein coding.
        
    Notes:
        1. `ensembl` object.
            from pyensembl import EnsemblRelease
            ensembl EnsemblRelease(release=100)    
    """    
    try:
        if geneid:
            g=ensembl.gene_by_id(k)
        else:
            g=ensembl.transcript_by_id(k)
    except:
        logging.error('gene id not found')
        return 
    return g.is_protein_coding

#restful api    
def rest(ids: list,
        function: str='lookup',
        target_taxon: str='9606',
        release: str='100',
        format_: str='full',
        test: bool=False,
        **kws):
    """Query Ensembl database using REST API.

    Args:
        ids (list): ids.
        function (str, optional): query function. Defaults to 'lookup'.
        target_taxon (str, optional): taxonomic id of the species. Defaults to '9606'.
        release (str, optional): ensembl release. Defaults to '100'.
        format_ (str, optional): format of the output. Defaults to 'full'.
        test (bool, optional): test mode. Defaults to False.

    Keyword Args:
        kws: additional queries.

    Raises:
        ValueError: ids should be str or list.

    Returns:
        dict: output.
    """
    import requests, sys

    server = f"https://e{release}.rest.ensembl.org"
    ext = f"/{function}/id"
    headers={ "Content-Type" : "application/json", "Accept" : "application/json"}
    
    headers['target_taxon']=target_taxon
    headers['release']=release
    headers['format']=format_
    headers.update(kws)
    if test: print(headers)
        
    if isinstance(ids,str):
        r = requests.get(server+ext+f'{ids}?', headers=headers)
    elif isinstance(ids,list):
        r = requests.post(server+ext, headers=headers, 
                          data='{ "ids" : ['+', '.join([f'"{s}"' for s in ids])+' ] }')
    else:
        raise ValueError(f"ids should be str or list")
    if not r.ok:
        r.raise_for_status()
    else:
        decoded = r.json()
    #     print(repr(decoded))
        return decoded

def to_homology(
                    x: str,
                    release: int=100,
                    homologytype: str='orthologues',
                    outd: str='data/database',
                    force: bool=False
                    ) -> dict:
    """
    Query homology of a gene using Ensembl REST API.
        
    Args:
        x (str): gene id.
        release (int, optional): Ensembl release number. Defaults to 100.
        homologytype (str, optional): type of the homology. Defaults to 'orthologues'.
        outd (str, optional): path of the output folder. Defaults to 'data/database'.
        force (bool, optional): overwrite output. Defaults to False.

    Returns:
        dict: output.

    References:
        1. Documentation: https://e{release}.rest.ensembl.org/documentation/info/homology_ensemblgene
    """
    p=f"https://e{release}.rest.ensembl.org/homology/id/{x}?type={homologytype};compara=vertebrates;sequence=none;cigar_line=0;content-type=application/json;format=full"
    outp=outp=f"{outd}/{p.replace('https://','')}.json"
    if exists(outp) and not force:
        return read_dict(outp)
    else:
        d1=read_dict(p)
        to_dict(d1,outp)
    return d1

def to_domains(x: str,
                    release: int,
                    species: str='homo_sapiens',
                    outd: str='data/database',
                    force: bool=False
                    ) -> pd.DataFrame:
    """Protein id to domains. 

    Args:
        x (str): protein id.
        release (int): Ensembl release.
        species (str, optional): species name. Defaults to 'homo_sapiens'.
        outd (str, optional): path of the output directory. Defaults to 'data/database'.
        force (bool, optional): overwrite output. Defaults to False.

    Returns:
        pd.DataFrame: output.
    """
    species=species.lower().replace(' ','_')
    p=f'https://e{release}.rest.ensembl.org/overlap/translation/{x}?content-type=application/json;species={species};feature=protein_feature;type=pfam'
    outp=outp=f"{outd}/{p.replace('https://','')}.json"
    if exists(outp) and not force:
        d1=read_dict(outp)
    else:
        d1=read_dict(p)
        to_dict(d1,outp)
    if d1 is None: return
    if len(d1)==0:
        logging.error(f"{x}: no domains/regions found")
        return
    #d1 is a list
    return pd.concat([pd.DataFrame(pd.Series(d)).T for d in d1],
                     axis=0)

## species
def to_species_name(k: str) -> str:
    """Convert to species name.

    Args:
        k (_type_): taxonomic id.

    Returns:
        str: species name.
    """
    server = f"https://e{release}.rest.ensembl.org"
    ext = f"/taxonomy/id/{k}?"
    r = requests.get(server+ext, headers={ "Content-Type" : "application/json"})
    if not r.ok:
        r.raise_for_status()
        sys.exit()
    decoded = r.json()
    return decoded['scientific_name']

def to_taxid(k: str) -> str:
    """Convert to taxonomic ids.  

    Args:
        k (str): species name.

    Returns:
        str: taxonomic id.
    """
    server = f"https://e{release}.rest.ensembl.org"
    ext = f"/taxonomy/name/{k}?"
    r = requests.get(server+ext, headers={ "Content-Type" : "application/json"})
    if not r.ok or r.status_code==400:
        logging.warning(f'no tax id found for {k}')
        return 
    decoded = r.json()
    if len(decoded)!=0:
        return decoded[0]['id']
    else:
        logging.warning(f'no tax id found for {k}')
        return
    
## convert between assemblies    
def convert_coords_human_assemblies(release: int,
                                    chrom: str,
                                    start: int,
                                    end: int,
                                    frm: int=38,
                                    to: int=37,
                                    test: bool=False,
                                    force: bool=False) -> dict:
    """Convert coordinates between human assemblies.

    Args:
        release (int): Ensembl release.
        chrom (str): chromosome name.
        start (int): start position.
        end (int): end position.
        frm (int, optional): assembly to convert from. Defaults to 38.
        to (int, optional): assembly to convert to. Defaults to 37.
        test (bool, optional): test mode. Defaults to False.
        force (bool, optional): overwrite outputs. Defaults to False.

    Returns:
        dict: output.
    """
    import requests, sys,yaml 
    server = f"https://e{release}.rest.ensembl.org"
    ext = f"/map/human/GRCh{frm}/{chrom}:{start}..{end}:1/GRCh{to}?"
    r = requests.get(server+ext, headers={ "Content-Type" : "application/json"})
    if test: logging.info(r.url)
    if not r.ok:
        if not force:
            r.raise_for_status()
        else:
            logging.info(f"not ok: {r.url}")
    return yaml.safe_load('\n'.join(r.text.split('\n')[3:-1]))
#     decoded = r.json()
#     d=eval(repr(decoded))
#     if 'mappings' in d:
#         for d_ in d['mappings']:
#             if 'mapped' in d_:
# #                 return d_['mapped']['seq_region_name'],d_['mapped']['start'],d_['mapped']['end']
#                 return pd.Series(d_['mapped'])#['seq_region_name'],d_['mapped']['start'],d_['mapped']['end']

def map_id(df1: pd.DataFrame,
            gene_id: str,
            release: str,
            release_to: str,
            out: str='df',
            test: bool=False) -> pd.DataFrame:
    """Map ids between releases.

    Args:
        df1 (pd.DataFrame): input dataframe.
        gene_id (str): gene id.
        release (str): release to convert from.
        release_to (str): release to convert to.
        out (str, optional): output type. Defaults to 'df'.
        test (bool, optional): test mode. Defaults to False.

    Returns:
        pd.DataFrame: output.

    Notes:     
        1. m:m mappings are possible. e.g. https://useast.ensembl.org/Homo_sapiens/Gene/Idhistory?db=core;g=ENSG00000276410;r=6:26043227-26043713;t=ENST00000615966
    """
    def get_release(df2,release,which,
                    col='Release',
                    test=False):
        # try:
        if test: print(df2)
        df1_=df2.loc[(df2[col]<=release),:]
        if test: print(df1_)
        if len(df1_)!=0:
            df_=df1_.tail(1)
        elif which=='old':
            df_=df2.head(1)
        elif which=='new':
            df_=df2.tail(1)
        if test: print(df_)
        return df_.iloc[-1,:][col]
        
    df2=df1.loc[df1['Old stable ID'].str.startswith(gene_id),:].sort_values('Release')
    if len(df2)==0:
        # mostly new genes that are uncharacterzed in the old release 
        # e.g. https://uswest.ensembl.org/Homo_sapiens/Gene/Idhistory?db=core;g=ENSG00000272104;r=3:50350892-50367923;t=ENST00000606589
        return [gene_id]
    df2_=df2.loc[(df2['Release']==get_release(df2,release=release,which='old',test=test)),['Old stable ID','Release']].drop_duplicates()
    if test: print(df2_)
    # try:
    assert(len(df2_)==1)
    df3=df2.loc[(df2['Release']==get_release(df2,release=release_to,which='new',test=test)),:].drop(['Old stable ID'],axis=1)
    df3['id']=df3['New stable ID'].str.split('.',expand=True)[0]
    df3['id']=df3['id'].replace('<retired>',np.nan)
    if out=='list':
        return df3['id'].dropna().unique().tolist()
    elif out=='df':
        ## more info
        df3['old id.version']=df2_.iloc[0,:]['Old stable ID']
        df3['old id updated on release']=df2_.iloc[0,:]['Release']
        df4=df3.rename(columns={'New stable ID':'id.version',
                           'Release':'id updated on release',
                           },errors='raise').rd.lower_columns()
        return df4
    # except:
    #     print(df2)
    #     return 'error'

def read_idmapper_output(outp: str) -> pd.DataFrame:
    """Read the output of Ensembl's idmapper.

    Args:
        outp (str): path to the file.

    Returns:
        pd.DataFrame: output.
    """
    from pathlib import Path
    file = Path(outp)
    file.write_text(file.read_text().replace('.Old stable ID', 'Old stable ID'))
    df01=pd.read_csv(outp,#+'_',
                     sep=', ')
    df1=df01.log('Old stable ID').loc[(df01['Old stable ID']!='Old stable ID'),:].log('Old stable ID')
    assert(df1['Old stable ID'].str.startswith('ENSG').all())
    df1['Release']=df1['Release'].astype(float)
    return df1

def map_ids_(ids: list,df00: pd.DataFrame,release: int,release_to: int) -> pd.DataFrame:
    """Function for mapping many ids.

    Args:
        ids (list): list of ids.
        df00 (pd.DataFrame): source dataframe.
        release (str): release to convert from.
        release_to (str): release to convert to.

    Returns:
        pd.DataFrame: output.
    """
    df0=pd.DataFrame(ids,columns=[f'id {release}'])
    df0['id']=df0[f'id {release}'].parallel_apply(lambda x: map_id(df00,gene_id=x,
                                                   release=release,
                                                   release_to=release_to,
                                                   out='list'))
    assert(not df0['id'].isnull().any())
    # reomve retired ids
    df2=df0.log().explode('id').log.dropna(subset=['id'])
    return df2

def map_ids(srcp: str,
        dbp: str,
        ids: list,
        release: int=75,
        release_to: int=100,
        species: str='human',
        test: bool=False) -> pd.DataFrame:
    """Map many ids between Ensembl releases.

    Args:
        srcp (str): path to the IDmapper.pl file.
        dbp (str): path to the database.
        ids (list): list of ids.
        release (str): release to convert from.
        release_to (str): release to convert to.
        species (str, optional): species name. Defaults to 'human'.
        test (bool, optional): test mode. Defaults to False.

    Returns:
        pd.DataFrame: output.

    Examples:
        srcp='deps/ensembl-tools/scripts/id_history_converter/IDmapper.pl',
        dbp='data/database/ensembl_id_history_converter/db.pqt',
        ids=ensembl.gene_ids(),    
    """
    from roux.query.ensembl import map_ids_,read_idmapper_output
    if exists(dbp):
        df00=read_table(dbp)
        df1=map_ids_(ids,df00,release,release_to)
        if test:print(df1.columns.tolist())
        ids_=list(set(ids) - set(df1[f'id {release}'].tolist()))
    else:
        logging.warning(f"database not found: {dbp}")
        ids_=ids
    if len(ids_)==0:
        logging.warning(f"ids not found in database, will convert: {len(ids_)}")
        return df1
    #run idmapper on the remaining
    outp=abspath('test/ids.out')
    import os
    if exists(outp):os.remove(outp)
    runbash_tmp(f"cd {dirname(srcp)};./{basename(srcp)} -s {species} -f INPUT > OUTPUT",
                env='ensembl',
                df1=ids_,
                input_type='list',
                inp='INPUT',outp=outp,
                force=True,
                test=test,
                # **kws,
               )
    if exists(outp):
        if test:info(outp)
        df2=read_idmapper_output(outp)
        if exists(dbp):
            df4=df00.log().append(df2).log()
            logging.warning(f"database updated: {dbp}")
        else:
            df4=df2.copy()
        to_table(df4,dbp)
        df3=map_ids_(ids_,df2,release,release_to)
    else:
        logging.warning("new ids could not be converted")
        if 'df1' in locals():
            df3=df1.copy()
        else:
            return
    return df3



"""
io_sys -> io_files
"""
import subprocess
import sys
from os.path import exists,dirname,basename,abspath,isdir,realpath,splitext ## prefer `pathlib` over `os.path`
from pathlib import Path
from glob import glob,iglob
import logging
# from roux.lib.io import makedirs
# from roux.global_imports import info

# walker
def get_all_subpaths(d='.',include_directories=False):
    """Get all the subpaths.

    Args:
        d (str, optional): _description_. Defaults to '.'.
        include_directories (bool, optional): to include the directories. Defaults to False.

    Returns:
        paths (list): sub-paths.
    """
    from glob import glob
    import os
    paths=[]
    for root, dirs, files in os.walk(d):
        if include_directories:
            for d in dirs:
                path=os.path.relpath(os.path.join(root, d), ".")
                paths.append(path)
        for f in files:
            path=os.path.relpath(os.path.join(root, f), d)
            paths.append(path)
    paths=sorted(paths)
    return paths

def basenamenoext(p):
    """Basename without the extension.

    Args:
        p (str): path.

    Returns:
        s (str): output.
    """
    return splitext(basename(p))[0]

def remove_exts(
    p: str,
    exts: tuple= None,
    ):
    """Filename without the extension.

    Args:
        p (str): path.
        exts (tuple): extensions.

    Returns:
        s (str): output.
    """
    return str(p).rstrip(''.join(Path(p).suffixes) if exts is None else exts)

def makedirs(p: str,exist_ok=True,**kws):
    """Make directories recursively.

    Args:
        p (str): path.
        exist_ok (bool, optional): no error if the directory exists. Defaults to True.

    Returns:
        p_ (str): the path of the directory.
    """
    from os import makedirs
    from os.path import isdir
    p_=p
    if not isdir(p):
        p=dirname(p)
    makedirs(p,exist_ok=exist_ok,**kws)
    return p_

def get_env(env_name: str):
    """Get the virtual environment as a dictionary.

    Args:
        env_name (str): name of the environment.

    Returns:
        d (dict): parameters of the virtual environment.
    """
    import sys,subprocess, os
    env = os.environ.copy()
    env_name_current=sys.executable.split('anaconda3/envs/')[1].split('/')[0]
    path=sys.executable.replace(env_name_current,env_name)
    env['CONDA_PYTHON_EXE']=path
    if 'anaconda3/envs' in env["PATH"]:
        env["PATH"]=env["PATH"].replace(env_name_current,env_name)
    elif 'anaconda' in env["PATH"]:
        env["PATH"]=env["PATH"].replace(f"{sys.executable.split('/anaconda3')[0]}/anaconda3/bin",
                                        f"{sys.executable.split('/anaconda3')[0]}/anaconda3/envs/{env_name}/bin")
    else:
        env["PATH"]=path.replace('/bin/python','/bin')+':'+env["PATH"]
    return env

def runbash(s1,env=None,test=False,**kws):
    """Run a bash command. 

    Args:
        s1 (str): command.
        env (str): environment name.
        test (bool, optional): testing. Defaults to False.

    Returns:
        output: output of the `subprocess.call` function.

    TODOs:
        1. logp
        2. error ignoring
    """
    if test:logging.info(s1)
    if env is None:
        logging.warning('env is not set.')
    response=subprocess.call(s1, shell=True,
                           env=get_env(env) if isinstance(env,str) else env if not env is None else env,
               stderr=subprocess.DEVNULL if not test else None, 
               stdout=subprocess.DEVNULL if not test else None,
               **kws)
    assert response==0, f"Error: {s1}"+('\nset `test=True` for more verbose.' if not test else '')
    return response

def runbash_tmp(s1: str,
            env: str,
            df1=None,
            inp='INPUT',
            input_type='df',
            output_type='path',
            tmp_infn='in.txt',
            tmp_outfn='out.txt',
            outp=None,
            force=False,
            test=False,
            **kws):
    """Run a bash command in `/tmp` directory.

    Args:
        s1 (str): command.
        env (str): environment name.
        df1 (DataFrame, optional): input dataframe. Defaults to None.
        inp (str, optional): input path. Defaults to 'INPUT'.
        input_type (str, optional): input type. Defaults to 'df'.
        output_type (str, optional): output type. Defaults to 'path'.
        tmp_infn (str, optional): temporary input file. Defaults to 'in.txt'.
        tmp_outfn (str, optional): temporary output file.. Defaults to 'out.txt'.
        outp (_type_, optional): output path. Defaults to None.
        force (bool, optional): force. Defaults to False.
        test (bool, optional): test. Defaults to False.

    Returns:
        output: output of the `subprocess.call` function.
    """
    if exists(outp) and not force:
        return
    from roux.lib.str import replace_many
    import shutil
    import tempfile
    with tempfile.TemporaryDirectory() as p:
        if test: p=abspath('test/')
        makedirs(p)
        tmp_inp=f"{p}/{tmp_infn}"
        tmp_outp=f"{p}/{tmp_outfn}"
        s1=replace_many(s1,{'INPUT':tmp_inp,
                            'OUTPUT':tmp_outp,
                           })
        if not df1 is None:
            if input_type=='df':
                df1.to_csv(replace_many(inp,{'INPUT':tmp_inp,}),sep='\t')
            elif input_type=='list':
                from roux.lib.set import to_list
                to_list(df1,replace_many(inp,{'INPUT':tmp_inp}))
        response=runbash(s1,env=env,
            test=test,
            **kws) 
        if exists(tmp_outp):
            if output_type=='path':
                makedirs(outp)
                shutil.move(tmp_outp,outp)
                return outp
        else:
            logging.error(f"output file not found: {outp} ({tmp_outp})")
            
def create_symlink(p: str,outp: str,test=False):
    """Create symbolic links.

    Args:
        p (str): input path.
        outp (str): output path.
        test (bool, optional): test. Defaults to False.

    Returns:
        outp (str): output path.
    """
    import os
    p,outp=abspath(p),abspath(outp)
    com=f"ln -s {p} {dirname(outp)}"
    if os.path.islink(outp):
        if os.readlink(outp)==abspath(p):
            return com
        else:
            logging.error(f"skipped: wrong symlink {os.readlink(outp)} not {outp}")
            return
    if exists(outp):
        logging.error(f"skipped: file exists {outp}")
        return
    if not exists(p):
        logging.error(f"skipped: file does not exists {p}")
        return 
    makedirs(abspath(dirname(outp)))
    if test: print(com)
    os.system(com)
    return outp

def input_binary(q:str):
    """Get input in binary format.

    Args:
        q (str): question.

    Returns:
        b (bool): response.
    """
    reply=''
    while not reply in ['y','n','o']:
        reply = input(f"{q}:")
        if reply == 'y':
            return True
        if reply == 'n':
            return False
    return reply

def is_interactive():
    """Check if the UI is interactive e.g. jupyter or command line. 
    """
    # thanks to https://stackoverflow.com/a/22424821/3521099
    import __main__ as main
    return not hasattr(main, '__file__')

def is_interactive_notebook():
    """Check if the UI is interactive e.g. jupyter or command line.     
    
    Notes:
        1. Difference in sys.module of notebook and shell
            'IPython.core.completerlib',
            'IPython.core.payloadpage',
            'IPython.utils.tokenutil',
            '_sysconfigdata_m_linux_x86_64-linux-gnu',
            'faulthandler',
            'imp',
            'ipykernel.codeutil',
            'ipykernel.datapub',
            'ipykernel.displayhook',
            'ipykernel.heartbeat',
            'ipykernel.iostream',
            'ipykernel.ipkernel',
            'ipykernel.kernelapp',
            'ipykernel.parentpoller',
            'ipykernel.pickleutil',
            'ipykernel.pylab',
            'ipykernel.pylab.backend_inline',
            'ipykernel.pylab.config',
            'ipykernel.serialize',
            'ipykernel.zmqshell',
            'storemagic'
        
        Code to find the difference:
            from roux.global_imports import *
            import sys
            with open('notebook.txt','w') as f:
                f.write('\n'.join(sys.modules))

            from roux.global_imports import *
            import sys
            with open('shell.txt','w') as f:
                f.write('\n'.join(sys.modules))
            set(open('notebook.txt','r').read().split('\n')).difference(open('shell.txt','r').read().split('\n'))    

    Reference:
        1. https://stackoverflow.com/a/22424821
    """
    return 'ipykernel.kernelapp' in sys.modules

def get_excecution_location(depth=1):
    """Get the location of the function being executed.

    Args:
        depth (int, optional): Depth of the location. Defaults to 1.

    Returns:
        tuple (tuple): filename and line number.
    """
    from inspect import getframeinfo, stack
    caller = getframeinfo(stack()[depth][0])
    return caller.filename,caller.lineno

## time
def get_time():
    """Gets current time in a form of a formated string. Used in logger function.
    """
    from roux.lib.io import to_path # potential circular import
    import datetime
    time=to_path('%s' % datetime.datetime.now())
    return time.replace('-','_').replace(':','_').replace('.','_')

## logging system
def get_datetime(outstr=True):
    """Get the date and time.

    Args:
        outstr (bool, optional): string output. Defaults to True.

    Returns:
        s : date and time.
    """
    from roux.lib.io import to_path # potential circular import
    import datetime
    time=datetime.datetime.now()
    if outstr:
        return to_path(str(time)).replace('-','_')
    else:
        return time

def p2time(filename: str,time_type='m'):
    """Get the creation/modification dates of files.

    Args:
        filename (str): filename.
        time_type (str, optional): _description_. Defaults to 'm'.

    Returns:
        time (str): time.
    """
    import os
    import datetime
    if time_type=='m':
        t = os.path.getmtime(filename)
    else:
        t = os.path.getctime(filename)
    return str(datetime.datetime.fromtimestamp(t))

def ps2time(ps: list,**kws_p2time):
    """Get the times for a list of files. 

    Args:
        ps (list): list of paths.

    Returns:
        ds (Series): paths mapped to corresponding times.
    """
    import pandas as pd
    from glob import glob
    if isinstance(ps,str):
        if isdir(ps):
            ps=glob(f"{ps}/*")
    return pd.Series({p:p2time(p,**kws_p2time) for p in ps}).sort_values().reset_index().rename(columns={'index':'p',0:'time'})
    

def get_logger(program='program',argv=None,level=None,dp=None):
    """Get the logging object.

    Args:
        program (str, optional): name of the program. Defaults to 'program'.
        argv (_type_, optional): arguments. Defaults to None.
        level (_type_, optional): level of logging. Defaults to None.
        dp (_type_, optional): _description_. Defaults to None.
    """
    log_format='[%(asctime)s] %(levelname)s\tfrom %(filename)s in %(funcName)s(..):%(lineno)d: %(message)s'
# def initialize_logger(output_dir):
    cmd='_'.join([str(s) for s in argv]).replace('/','_')
    if dp is None:
        dp=''
    else:
        dp=dp+'/'
    date=get_datetime()
    logp=f"{dp}.log_{program}_{date}_{cmd}.log"
    #'[%(asctime)s] %(levelname)s\tfrom %(filename)s in %(funcName)s(..):%(lineno)d: %(message)s'
    
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    
    # create console handler and set level to info
    handler = logging.StreamHandler()
    handler.setLevel(level)
    formatter = logging.Formatter(log_format)
    handler.setFormatter(formatter)
    logger.addHandler(handler)

#     # create error file handler and set level to error
#     handler = logging.FileHandler(os.path.join(output_dir, "error.log"),"w", encoding=None, delay="true")
#     handler.setLevel(logging.ERROR)
#     formatter = logging.Formatter(log_format)
#     handler.setFormatter(formatter)
#     logger.addHandler(handler)

    # create debug file handler and set level to debug
    handler = logging.FileHandler(logp)
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter(log_format)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logp

# log
from icecream import ic as info
info.configureOutput(prefix='INFO:icrm:')


# import pandas as pd
from roux.global_imports import *

def get_paired_sets_stats(l1: list,l2: list,test: bool=False) -> list:
    """Paired stats comparing two sets.

    Args:
        l1 (list): set #1.
        l2 (list): set #2.
        test (bool): test mode. Defaults to False.

    Returns:
        list: tuple (overlap, intersection, union, ratio).
    """
    if all([isinstance(l, list) for l  in [l1,l2]]):
        if test: print(l1,l2)
        l=list(jaccard_index(l1,l2))
        from roux.stat.diff import get_ratio_sorted
        l.append(get_ratio_sorted(len(l1),len(l2)))
        return l

def get_stats_paired(df1: pd.DataFrame,
                    cols: list,
                    input_logscale: bool,
                     prefix: str=None,
                     drop_cols: bool=False,
                     unidirectional_stats: list=['min','max'],
                     fast: bool=False) -> pd.DataFrame:
    """Paired stats, row-wise.

    Args:
        df1 (pd.DataFrame): input data.
        cols (list): columns.
        input_logscale (bool): if the input data is log-scaled.
        prefix (str, optional): prefix of the output column/s. Defaults to None.
        drop_cols (bool, optional): drop these columns. Defaults to False.
        unidirectional_stats (list, optional): column-wise status. Defaults to ['min','max'].
        fast (bool, optional): parallel processing. Defaults to False.

    Returns:
        pd.DataFrame: output dataframe.
    """
    assert(len(cols)==2)
    if prefix is None:
        prefix=get_fix(*cols,common=True,clean=True)
        info(prefix)
    from roux.stat.diff import get_ratio_sorted,get_diff_sorted
    df1[f"{prefix} {'ratio' if not input_logscale else 'diff'}"]=getattr(
        df1,'parallel_apply' if fast else "apply"
        )(lambda x: (
            get_ratio_sorted if not input_logscale else get_diff_sorted
        )(x[cols[0]],x[cols[1]]),axis=1
        )
    assert(not any(df1[f"{prefix} {'ratio' if not input_logscale else 'diff'}"]<0))
    for k in unidirectional_stats:
        df1[f'{prefix} {k}']=getattr(df1.loc[:,cols],k)(axis=1)
    if drop_cols:
        df1=df1.log.drop(labels=cols,axis=1)
    return df1

def get_stats_paired_agg(x: np.array,y: np.array,ignore: bool=False,verb: bool=True) -> pd.Series:
    """Paired stats aggregated, for example, to classify 2D distributions.

    Args:
        x (np.array): x vector.
        y (np.array): y vector.
        ignore (bool, optional): suppress warnings. Defaults to False.
        verb (bool, optional): verbose. Defaults to True.

    Returns:
        pd.Series: output.
    """
    ## escape duplicate value inputs
    if not ignore:
        if len(x)!=len(y):
            if verb: logging.error("len(x)!=len(y)")
            return
        if len(x)<5:
            if verb: logging.error("at least 5 data points needed.")
            return
        if ((x.nunique()/len(x))<0.5) or ((y.nunique()/len(y))<0.5):
            if verb: logging.error("half or more duplicate values.")        
            return
    d={}
    d['n']=len(x)
    from roux.stat.corr import get_spearmanr,get_pearsonr
    d['$r_s$'],d['P ($r_s$)']=get_spearmanr(x,y)
    d['linear regression slope'],d['linear regression y-intercept'],d['$r_p$'],d['P ($r_p$)'],d['linear regression stderr']=sc.stats.linregress(x,y)
#     shape of distributions
    for k,a in zip(['x','y'],[x,y]):
        from roux.stat.cluster import cluster_1d
        d2=cluster_1d(ds=a,
                   n_clusters=2,
                    returns=['clf'],             
                       clf_type='gmm',
                   random_state=88,
#                    test=True,
                        )    
        d[f'{k} peak1 weight'],d[f'{k} peak2 weight'] = d2['clf'].weights_.flatten()
        d[f'{k} peak1 mean'],d[f'{k} peak2 mean'] = d2['clf'].means_.flatten()
        d[f'{k} peak1 std'],d[f'{k} peak2 std']=np.sqrt(d2['clf'].covariances_).ravel().reshape(2,1).flatten()    
    return pd.Series(d)    


from roux.lib.io import *
from roux.query.ensembl import release2prefix

def get_ensembl_dataset_name(x: str) -> str:
    """Get the name of the Ensembl dataset.

    Args:
        x (str): species name.

    Returns:
        str: output.
    """
    l=x.lower().split(' ')
    assert(len(l)==2)
    return f"{l[0][0]}{l[1]}_gene_ensembl"

def query(
          species: str,
          release: int,
          attributes: list=None,
          filters: list=None,
          databasep: str='data/database',
          dataset_name: str=None,
          force: bool=False,
          **kws_query,) -> pd.DataFrame:
    """Query the biomart database.

    Args:
        species (str): species name.
        release (int): Ensembl release. 
        attributes (list, optional): list of attributes. Defaults to None.
        filters (list, optional): list of filters. Defaults to None.
        databasep (str, optional): path to the local database folder. Defaults to 'data/database'.
        dataset_name (str, optional): dataset name. Defaults to None.
        force (bool, optional): overwrite output. Defaults to False.

    Returns:
        pd.DataFrame: output

    Examples: 
        1. Setting filters for the human data:
            filters={
            # REMOVE: mitochondria/sex chr
                'chromosome_name':[str(i) for i in list(range(1,23))],
            # REMOVE: non protein coding
                'biotype':['protein_coding'],
                }
 
    """
    if dataset_name is None:
        dataset_name=get_ensembl_dataset_name(species)
    from pybiomart import Server,Dataset
    serverp=f"http://{release2prefix[release]}.ensembl.org"       
    server = Server(host=serverp)
    assert release==int(server['ENSEMBL_MART_ENSEMBL'].display_name.split(' ')[-1]), server['ENSEMBL_MART_ENSEMBL'].display_name
#     release=server['ENSEMBL_MART_ENSEMBL'].display_name.split(' ')[-1]
    logging.info(f"{dataset_name} version: {release} is used")
    dataset = Dataset(name=dataset_name,host=serverp)
    if attributes is None:
        to_table(dataset.list_attributes(),'test/biomart_attributes.tsv')
        logging.info("choose the attributes from: test/biomart_attributes.tsv")
        attributes = input(f"attributes space separated. e.g. a,b:").split(' ')
    if filters is None:
        to_table(dataset.list_filters(),'test/biomart_filters.tsv')        
        logging.info("choose the attributes from: test/biomart_filters.tsv")
        filters = eval(input(f"filters as python dict."))
    outp=f"{databasep}/www.ensembl.org/biomart/{dataset_name}/{release}/{'_'.join(sorted(attributes))}_{'_'.join(sorted(list(filters.keys())))}.pqt"
    if len(basenamenoext(outp))>50:
        outp=f"{dirname(outp)}/{basenamenoext(outp)[:50]}_plus{len(basenamenoext(outp))-50}{splitext(outp)[1]}"
    if not exists(outp) or force:
        df1=dataset.query(attributes=np.unique(attributes),
                         filters=filters,
                         only_unique=True,
                         **kws_query)
        to_table(df1,outp)
    else:
        df1=read_table(outp)
    return df1


from roux.global_imports import *

def fit_curve_fit(func,
                  xdata: np.array = None,
                  ydata: np.array=None,
                  bounds: tuple =(-np.inf, np.inf),
                  test=False,
                  plot=False) -> tuple:
    """Wrapper around `scipy`'s `curve_fit`.

    Args:
        func (function): fitting function.
        xdata (np.array, optional): x data. Defaults to None.
        ydata (np.array, optional): y data. Defaults to None.
        bounds (tuple, optional): bounds. Defaults to (-np.inf, np.inf).
        test (bool, optional): test. Defaults to False.
        plot (bool, optional): plot. Defaults to False.

    Returns:
        tuple: output.
    """
    from scipy.optimize import curve_fit
    # >>>
    # Define the data to be fit with some noise:
    if xdata is None and ydata is None:
        # >>>
        xdata = np.linspace(1, 4, 50)
        y = func(xdata, -2.5)
        np.random.seed(1729)
        y_noise = 0.2 * np.random.normal(size=xdata.size)
        ydata = y + y_noise
    if test or plot:
        plt.plot(xdata, ydata, 'b.', label='data')
        # Fit for the parameters a, b, c of the function func:

    # >>>
    popt, pcov = curve_fit(func, xdata, ydata)
    if test or plot:
        print(popt)
        plt.plot(xdata, func(xdata, *popt), 'r-',
                 label=f'non-bounded fit:\na={popt[0]}')
    # Constrain the optimization to the region of 0 <= a <= 3, 0 <= b <= 1 and 0 <= c <= 0.5:
    # >>>
    popt, pcov = curve_fit(func, xdata, ydata, bounds=bounds)
    
    if test or plot:
        print(popt)
        plt.plot(xdata, func(xdata, *popt), 'g--',
                 label=f'bounded fit:\na={popt[0]}')
        # >>>
        plt.xlabel('x')
        plt.ylabel('y')
        plt.legend()
        plt.show()
    return func(xdata, *popt),popt

def fit_gauss_bimodal(
    data: np.array,
    bins: int=50,
    expected: tuple=(1,.2,250,2,.2,125),
    test=False) -> tuple:
    """Fit bimodal gaussian distribution to the data in vector format.

    Args:
        data (np.array): vector.
        bins (int, optional): bins. Defaults to 50.
        expected (tuple, optional): expected parameters. Defaults to (1,.2,250,2,.2,125).
        test (bool, optional): test. Defaults to False.

    Returns:
        tuple: _description_

    Notes:
        Observed better performance with `roux.stat.cluster.cluster_1d`.
    """
    from scipy.optimize import curve_fit
    def gauss(x,mu,sigma,A):
        return A*np.exp(-(x-mu)**2/2/sigma**2)
    def bimodal(x,mu1,sigma1,A1,mu2,sigma2,A2):
        return gauss(x,mu1,sigma1,A1)+gauss(x,mu2,sigma2,A2)
    y,x=np.histogram(data, bins=bins,density=True)
    x=(x[:-1] + x[1:]) / 2
    params,cov=curve_fit(bimodal,x,y,expected)
    sigma=np.sqrt(np.diag(cov))
    if test:
        plt.figure()
        _=plt.hist(data,bins,alpha=.3,label='data',density=True)
        plt.plot(x,bimodal(x,*params),color='red',lw=3,label='model')
        plt.legend()
    return params,sigma

## 2D
def get_grid(x: np.array,
        y: np.array,
        z: np.array =None,
        off: int=0,
        grids: int=100,
        method='linear',
        test=False,
        **kws) -> tuple:
    """2D grids from 1d data.

    Args:
        x (np.array): vector.
        y (np.array): vector.
        z (np.array, optional): vector. Defaults to None.
        off (int, optional): offsets. Defaults to 0.
        grids (int, optional): grids. Defaults to 100.
        method (str, optional): method. Defaults to 'linear'.
        test (bool, optional): test. Defaults to False.

    Returns:
        tuple: output.
    """
    xoff=(np.max(x)-np.min(x))*off
    yoff=(np.max(y)-np.min(y))*off    
    xi=np.linspace(np.min(x)-xoff,np.max(x)+xoff,grids)
    yi=np.linspace(np.min(y)-yoff,np.max(y)+yoff,grids)

    X,Y= np.meshgrid(xi,yi)
    if z is None:
        return X,Y
    else:
        Z = sc.interpolate.griddata((x, y), z, (X, Y),
                                    method=method,
                                    fill_value=min(z),
                                   )
        return X,Y,Z
    
def fit_gaussian2d(
    x: np.array,
    y: np.array,
    z: np.array,
    grid=True,
    grids=20,
    method='linear',
    off=0,
    rescalez=True,
    test=False) -> tuple:
    """Fit gaussian 2D.

    Args:
        x (np.array): vector.
        y (np.array): vector.
        z (np.array): vector.
        grid (bool, optional): grid. Defaults to True.
        grids (int, optional): grids. Defaults to 20.
        method (str, optional): method. Defaults to 'linear'.
        off (int, optional): offsets. Defaults to 0.
        rescalez (bool, optional): rescalez. Defaults to True.
        test (bool, optional): test. Defaults to False.

    Returns:
        tuple: output.
    """
    if grid:
        xg,yg,zg=get_grid(x,y,z,
                        grids=grids,
                        method=method)    
    else:
        xg,yg,zg=x,y,z
    from astropy.modeling import models,fitting
    if rescalez:
        from roux.stat.transform import rescale
        range1=(np.min(zg),np.max(zg))
        zg_=rescale(a=zg, range1=range1, range2=[0, 1])
        
#         from sklearn.preprocessing import MinMaxScaler
#         scaler = MinMaxScaler()
#         scaler.fit(zg)
#         zg_ = scaler.transform(zg)
        if test:
            print(np.min(zg),np.max(zg),zg.shape)        
            print(np.min(zg_),np.max(zg_),zg_.shape)                
    else:
        zg_=zg
        
    m1 = models.Gaussian2D(
        amplitude=np.max(zg_),
        x_mean=np.mean(xg), y_mean=np.mean(yg), 
#       x_stddev=np.std(xg), y_stddev=np.std(yg), 
#       theta=0.
        cov_matrix=np.cov(np.vstack([xg.flatten(), yg.flatten()])),
        fixed={'x_mean':True,'y_mean':True},
        bounds={'amplitude':(np.min(zg_),np.max(zg_))}
    )
#     z = m1(x, y)
    fitr = fitting.LevMarLSQFitter()
    mf1 = fitr(m1,xg,yg,zg_)
    if grid and off!=0:
        xg,yg=get_grid(x,y,
                        grids=grids,
                        method=method,
                      off=off)        
    zg_hat_=mf1(xg,yg)
    print(mf1.theta)
    if rescalez:
        zg_hat=rescale(a=zg_hat_, range1=[0,1], range2=range1)    
        # for inverse transformation
#         zg_hat = scaler.inverse_transform(zg_hat_)
        if test:
            print(np.min(zg_hat_),np.max(zg_hat_),zg_hat_.shape)        
            print(np.min(zg_hat),np.max(zg_hat),zg_hat.shape)        
#         print(range1)
    else:
        zg_hat=zg_hat_
    return xg,yg,zg,zg_hat  

def fit_2d_distribution_kde(x: np.array, 
                            y: np.array,
                            bandwidth: float, 
                            xmin: float=None,
                            xmax: float=None,
                            xbins=100j, 
                            ymin: float=None,
                            ymax: float=None,
                            ybins=100j, 
                            test=False,
                            **kwargs,) -> tuple: 
    """
    2D kernel density estimate (KDE).

    Notes:
        Cut off outliers:
            quantile_coff=0.01
            params_grid=merge_dicts([
            df01.loc[:,var2col.values()].quantile(quantile_coff).rename(index=flip_dict({f"{k}min":var2col[k] for k in var2col})).to_dict(),
            df01.loc[:,var2col.values()].quantile(1-quantile_coff).rename(index=flip_dict({f"{k}max":var2col[k] for k in var2col})).to_dict(),
                    ])

    Args:
        x (np.array): vector.
        y (np.array): vector.
        bandwidth (float): bandwidth
        xmin (float, optional): x minimum. Defaults to None.
        xmax (float, optional): x maximum. Defaults to None.
        xbins (_type_, optional): x bins. Defaults to 100j.
        ymin (float, optional): y minimum. Defaults to None.
        ymax (float, optional): y maximum. Defaults to None.
        ybins (_type_, optional): y bins. Defaults to 100j.
        test (bool, optional): test. Defaults to False.

    Returns:
        tuple: output.
    """
    from sklearn.neighbors import KernelDensity
    # create grid of sample locations (default: 100x100)
    xx, yy = np.mgrid[x.min() if xmin is None else xmin:x.max() if xmax is None else xmax:xbins, 
                      y.min() if ymin is None else ymin:y.max() if ymax is None else ymax:ybins]

    xy_sample = np.vstack([yy.ravel(), xx.ravel()]).T
    xy_train  = np.vstack([y, x]).T

    kde_skl = KernelDensity(bandwidth=bandwidth, **kwargs)
    kde_skl.fit(xy_train)

    # score_samples() returns the log-likelihood of the samples
    z = np.exp(kde_skl.score_samples(xy_sample))
    zz=np.reshape(z, xx.shape)
    if test:
        fig=plt.figure(figsize=[5,4])
        ax=plt.subplot()
        ax.scatter(x, y, s=1, fc='k')
        pc=ax.pcolormesh(xx, yy, zz,cmap='Reds')
        fig.colorbar(pc)
        ax.set(**{'xlim':[xmin,xmax],'ylim':[ymin,ymax],
                 'title':f"bandwidth{bandwidth}_bins{xbins}"})
    return xx, yy, zz    

def check_poly_fit(d: pd.DataFrame,
            xcol: str,
            ycol: str,
            degmax: int=5) -> pd.DataFrame:
    """Check the fit of a polynomial equations.

    Args:
        d (pd.DataFrame): input dataframe.
        xcol (str): column containing the x values.
        ycol (str): column containing the y values.
        degmax (int, optional): degree maximum. Defaults to 5.

    Returns:
        pd.DataFrame: _description_
    """
    from scipy.stats import linregress
    ns=range(1,degmax+1,1)
    plt.figure(figsize=[9,5])
    ax=plt.subplot(121)
    d.plot.scatter(x=xcol,y=ycol,alpha=0.3,color='gray',ax=ax)
    metrics=pd.DataFrame(index=ns,columns=['r','standard error'])
    for n in ns:
        fit = np.polyfit(d[xcol], d[ycol], n) 
        yp = np.poly1d(fit)
        _,_,r,_,e = linregress(yp(d[xcol]), d[ycol])
        metrics.loc[n,'r']=r
        metrics.loc[n,'standard error']=e    
        ax.plot(d[xcol], yp(d[xcol]), '-',color=plt.get_cmap('hsv')(n/float(max(ns))),
                alpha=1,
                lw='4')
    ax.legend(['degree=%d' % n for n in ns],bbox_to_anchor=(1, 1))

    metrics.index.name='degree'
    ax=plt.subplot(122)
    ax=metrics.plot.barh(ax=ax)
    ax.legend(bbox_to_anchor=[1,1])
#     metrics.plot.barh('standard error')
    plt.tight_layout()
    return metrics

def mlr_2(df: pd.DataFrame,coly: str,colxs: list) -> tuple:
    """Multiple linear regression between two variables.

    Args:
        df (pd.DataFrame): input dataframe.
        coly (str): column  containing y values.
        colxs (list): columns containing x values.

    Returns:
        tuple: output.
    """
    from sklearn.preprocessing import PolynomialFeatures
    from sklearn.linear_model import LinearRegression
    poly = PolynomialFeatures(interaction_only=True,include_bias = False)
    X=poly.fit_transform(df.loc[:,colxs])
    # pd.DataFrame()
    y=df.loc[:,coly].tolist()
    reg = LinearRegression().fit(X, y)
    label_score=f"$r^2$={reg.score(X, y):.2g}"
    label_eqn="$y$="+''.join([f"{l[0]:+.2g}*{l[1]}" for l in zip(reg.coef_,['$x_1$','$x_2$','$x_1$*$x_2$'])])[1:]
    dplot=pd.DataFrame({f'{coly}': y,
    f'{coly} predicted': reg.predict(X),})
    return label_score,label_eqn,dplot

def get_mlr_2_str(df: pd.DataFrame,coly: str,colxs: list) -> str:
    """Get the result of the multiple linear regression between two variables as a string.

    Args:
        df (pd.DataFrame): input dataframe.
        coly (str): column  containing y values.
        colxs (list): columns containing x values.

    Returns:
        str: output.
    """
    label_score,label_eqn,dplot_reg=mlr_2(df,coly,colxs)
    label_eqn=label_eqn.replace('$y$','$z$').replace('$x_1$','$x$').replace('$x_2$','$y$')
    return f"{label_eqn} ({label_score})"


import pandas as pd
from roux.lib.sys import *
from roux.lib.io import read_ps
from roux.lib.set import flatten
from roux.lib.dict import read_dict,is_dict

def clear_variables(
    dtype=None,
    variables=None,
    # globals_dict=None,
    ):
    """Clear dataframes from the workspace."""
    assert (dtype is None or variables is None) and (not (dtype is None and variables is None))
    import gc
    if variables is None:
        variables=[k for k in globals() if isinstance(globals()[k], dtype) and not k.startswith('_')]
    if len(variables)==0: return
    for k in variables:
        del globals()[k]
    gc.collect()
    logging.info(f"cleared {len(variables)} variables: {variables}.")

def clear_dataframes():
    return clear_variables(
    dtype=pd.DataFrame,
    variables=None,
    )

def get_lines(
    p: str,
    keep_comments: bool=True
    ) -> list:
    """Get lines of code from notebook.

    Args:
        p (str): path to notebook.
        keep_comments (bool, optional): keep comments. Defaults to True.

    Returns:
        list: lines.
    """
    import nbformat
    from nbconvert import PythonExporter
    import os
    if os.path.islink(p):
        p=os.readlink(p)
    nb = nbformat.read(p, nbformat.NO_CONVERT)
    exporter = PythonExporter()
    source, meta = exporter.from_notebook_node(nb)
    lines=source.split('\n')
    lines=[s for s in lines if (isinstance(s,str) and s!='' and len(s)<1000)]
    if not keep_comments:
        lines=[s for s in lines if not s.startswith('#')]
    return lines

def to_py(
    notebookp: str,
    pyp: str=None,
    force: bool=False,
    **kws_get_lines,
    ) -> str:
    """To python script (.py).

    Args:
        notebookp (str): path to the notebook path.
        pyp (str, optional): path to the python file. Defaults to None.
        force (bool, optional): overwrite output. Defaults to False.

    Returns:
        str: path of the output.
    """
    if pyp is None: pyp=notebookp.replace('.ipynb','.py')
    if exists(pyp) and not force: return 
    makedirs(pyp)
    l1=get_lines(notebookp, **kws_get_lines)
    l1='\n'.join(l1).encode('ascii', 'ignore').decode('ascii')
    with open(pyp, 'w+') as fh:
        fh.writelines(l1)
    return pyp

def import_from_file(
    pyp: str
    ):
    """Import functions from python (`.py`) file.

    Args:
        pyp (str): python file (`.py`).

    """
    from importlib.machinery import SourceFileLoader
    return SourceFileLoader(abspath(pyp), abspath(pyp)).load_module()

def to_parameters(
    f: object,
    test: bool=False
    ) -> dict:
    """Get function to parameters map.

    Args:
        f (object): function.
        test (bool, optional): test mode. Defaults to False.

    Returns:
        dict: output.
    """
    sign=inspect.signature(f)
    params={}
    for arg in sign.parameters:
        argo=sign.parameters[arg]
        params[argo.name]=argo.default
    #     break
    return params

def read_nb_md(
    p: str
    ) -> list:
    """Read notebook's documentation in the markdown cells.

    Args:
        p (str): path of the notebook.

    Returns:
        list: lines of the strings.
    """
    import nbformat
    from sys import argv
    l1=[]
    l1.append("# "+basenamenoext(p))
    nb = nbformat.read(p, nbformat.NO_CONVERT)
    l1+=[cell.source for cell in nb.cells if cell.cell_type == 'markdown']
    return l1

## metadata-related
def read_metadata(
    ind: str='metadata'
    ) -> dict:
    """Read metadata.

    Args:
        ind (str, optional): input directory. Defaults to 'metadata'.

    Returns:
        dict: output.

    TODOs:
        1. Metadata files include colors.yaml, database.yaml, constants.yaml etc.
    """
    p=f'{ind}/metadata.yaml'
    for p_ in [ind,p]:
        if not exists(p_):
            logging.warning(f'not found: {ind}')
            return 
    d1=read_dict(p)
    ## read dicts
    for k in d1:
        if isinstance(d1[k],list):
            if len(d1[k])<10:
                d_={}
                for p in d1[k]:
                    if isinstance(p,str):
                        if is_dict(p) and exists(p):
                            d_[basenamenoext(p)]=read_dict(p)
                        else:
                            d_[basenamenoext(p)]=p
                            logging.error(f"file not found: {p}")
                if len(d_)!=0:
                    d1[k]=d_
    for p_ in glob(f"{ind}/*"):
        if isdir(p_):
            if len(glob(f'{p_}/*.json'))!=0:
                if not basename(p_) in d1 and len(glob(f'{p_}/*.json'))!=0:
                    d1[basename(p_)]=read_dict(f'{p_}/*.json')
                elif isinstance(d1[basename(p_)],dict) and len(glob(f'{p_}/*.json'))!=0:
                    d1[basename(p_)].update(read_dict(f'{p_}/*.json'))
                else:
                    logging.warning(f"entry collision, could not include '{p_}/*.json'")
        else:
            if is_dict(p_) and exists(p_):
                d1[basenamenoext(p_)]=read_dict(p_)
            else:
                logging.error(f"file not found: {p_}")                
    logging.info(f"metadata read from {p} (+"+str(len(glob(f'{ind}/*.json')))+" jsons)")
    return d1

## create documentation
def to_info(
    p: str='*_*_v*.ipynb',
    outp: str='README.md') -> str:
    """Save README.md file.

    Args:
        p (str, optional): path of the notebook files that would be converted to "tasks". Defaults to '*_*_v*.ipynb'.
        outp (str, optional): path of the output file. Defaults to 'README.md'.

    Returns:
        str: path of the output file.
    """
    ps=read_ps(p)
    l1=flatten([read_nb_md(p) for p in ps])
    with open(outp,'w') as f:
        f.writelines([f"{s}\n" for s in l1])
    return outp

def make_symlinks(
    d1: dict,
    d2: dict,
    project_path: str,
    data: bool=True,
    notebook_suffix: str='_v',
    test: bool=False
    )->list:
    """Make symbolic links.

    Args: 
        d1 (dict): `project name` to `repo name`.
        d2 (dict): `task name` to tuple containing `from project name` `to project name`.
        project_path (str): path of the repository.
        data (bool, optional): make links for the data. Defaults to True.
        notebook_suffix (str, optional): suffix of the notebook file to be considered as a "task".
        test (bool, optional): test mode. Defaults to False.

    Returns:
        list: list of commands.
    """
    coms=[]
    for k in d2:
        ## notebook
        p=read_ps(f"{project_path}/{d2[k][0]}/code/{d1[d2[k][0]]}/{d1[d2[k][0]]}/{k.split('/')[0]}*{notebook_suffix}*.ipynb")[0]
        if test: print(p)
        outp=f"{project_path}/{d2[k][1]}/code/{d1[d2[k][1]]}/{d1[d2[k][1]]}/{basename(p)}"
        if test: print(outp)
        coms.append(create_symlink(p,outp))
        if data:
            ## data_analysed
            p=f"{project_path}/{d2[k][0]}/data/data_analysed/data{k}"
            outp=f"{project_path}/{d2[k][1]}/data/data_analysed/data{k}"
            coms.append(create_symlink(p,outp))
        # break
    return coms

def to_workflow(
    df2: pd.DataFrame,
    workflowp: str,
    tab: str='    '
    ) -> str:
    """Save workflow file.

    Args:
        df2 (pd.DataFrame): input table.
        workflowp (str): path of the workflow file.
        tab (str, optional): tab format. Defaults to '    '.

    Returns:
        str: path of the workflow file.
    """
    makedirs(workflowp)
    from roux.lib.set import list2str
    with open(workflowp,'w') as f:
        ## add rule all
        f.write("from roux.lib.dict import read_dict\nfrom roux.workflow.io import read_metadata\nmetadata=read_metadata()\n"
                +'report: "workflow/report_template.rst"\n'
                +"\nrule all:\n"
                 f"{tab}input:\n"
                 f"{tab}{tab}"
#                     +f",\n{tab}{tab}".join(flatten([flatten(l) for l in df2['output paths'].dropna().tolist()]))
                +f",\n{tab}{tab}".join(df2['output paths'].dropna().tolist())
                +"\n# rules below\n\n"
                +'\n'.join(df2['rule code'].dropna().tolist())\
               )
    return workflowp

def create_workflow_report(
    workflowp: str,
    env: str,
    ) -> int:
    """
    Create report for the workflow run.

    Parameters:
        workflowp (str): path of the workflow file (`snakemake`).
        env (str): name of the conda virtual environment where required the workflow dependency is available i.e. `snakemake`.
    """
    from pathlib import Path
    workflowdp=str(Path(workflowp).absolute().with_suffix(''))+'/'
    ## create a template file for the report
    report_templatep=Path(f"{workflowdp}/report_template.rst")
    if not report_templatep.exists():
        report_templatep.parents[0].mkdir(parents=True, exist_ok=True)
        report_templatep.touch()

    from roux.lib.sys import runbash
    runbash(f"snakemake --snakefile {workflowp} --rulegraph > {workflowdp}/workflow.dot;sed -i '/digraph/,$!d' {workflowdp}/workflow.dot",env=env)
    
    ## format the flow chart
    from roux.lib.set import read_list,to_list
    to_list([s.replace('task','').replace('_step','\n') for s in read_list(f'{workflowdp}/workflow.dot')],
            f'{workflowdp}/workflow.dot')
    
    runbash(f"dot -Tpng {workflowdp}/workflow.dot > {workflowdp}/workflow.png",env=env)
    runbash(f"snakemake -s workflow.py --report {workflowdp}/report.html",env=env)

def to_diff_notebooks(
    notebook_paths,
    url_prefix="https://localhost:8888/nbdime/difftool?",
    remove_prefix='file://', # for bash
    verbose=True,
    ) -> list:
    """
    "Diff" notebooks using `nbdiff` (https://nbdime.readthedocs.io/en/latest/)
    
    Todos:
        1. Deprecate if functionality added to `nbdiff-web`.
    """
    import itertools
    logging.warning('to_diff_notebooks is under development.')
    urls_input=[Path(p).absolute().as_uri() for p in notebook_paths]
    urls_output=[]
    for url_base,url_remote in list(itertools.product(urls_input[:1],urls_input[1:])):
        urls_output.append(f"{url_prefix}base={url_base.replace('file://','')}&remote={url_remote.replace('file://','')}")
    if verbose:
        logging.info('Differences between notebooks:')
        logging.info('\n'.join(urls_output))
    return urls_output